(*
Author: Robert Sison
*)

theory Example_CDDC_RISCLanguage_OnePlaceBuf
imports Example_CDDC_WhileLockLanguage_OnePlaceBuf
begin

(* ('Reg, 'Wrd) Mem *)
definition regs\<^sub>0 :: "(reg, 8 word) Mem"
where
  "regs\<^sub>0 \<equiv> \<lambda>_. 0"

abbreviation comp_outputs_OverlayDriver
where
  "comp_outputs_OverlayDriver \<equiv>
   let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_OverlayDriver
      in (PCs, l', nl', C', regs\<^sub>0)"

abbreviation comp_outputs_HidInputDriver
where
  "comp_outputs_HidInputDriver \<equiv>
   let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_HidInputDriver
      in (PCs, l', nl', C', regs\<^sub>0)"

abbreviation comp_outputs_Switch
where
  "comp_outputs_Switch \<equiv>
   let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_Switch
      in (PCs, l', nl', C', regs\<^sub>0)"

definition prog_OverlayDriver_RISC
where
  "prog_OverlayDriver_RISC \<equiv> map fst (fst comp_outputs_OverlayDriver)"

definition prog_OverlayDriver_Cs
where
  "prog_OverlayDriver_Cs \<equiv> map snd (fst comp_outputs_OverlayDriver)"

definition prog_HidInputDriver_RISC
where
  "prog_HidInputDriver_RISC \<equiv> map fst (fst comp_outputs_HidInputDriver)"

definition prog_HidInputDriver_Cs
where
  "prog_HidInputDriver_Cs \<equiv> map snd (fst comp_outputs_HidInputDriver)"

definition prog_Switch_RISC
where
  "prog_Switch_RISC \<equiv> map fst (fst comp_outputs_Switch)"

definition prog_Switch_Cs
where
  "prog_Switch_Cs \<equiv> map snd (fst comp_outputs_Switch)"

abbreviation comp_outputs
where
  "comp_outputs \<equiv> [comp_outputs_OverlayDriver, comp_outputs_HidInputDriver, comp_outputs_Switch]"

definition init_cms_CDDC_RISC
where
  "init_cms_CDDC_RISC \<equiv> [(((0, prog_OverlayDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0),
                          (((0, prog_HidInputDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0),
                          (((0, prog_Switch_RISC), regs\<^sub>0), RC.mds\<^sub>0)]"

lemma prog_OverlayDriver_input_reqs:
  "RC.compile_cmd_input_reqs RC_C\<^sub>0 None 0 prog_OverlayDriver"
  unfolding RC.compile_cmd_input_reqs_def RC.C\<^sub>0_def prog_OverlayDriver_def RC.no_unstable_exprs_def
  by (force simp:RC.regrec_stable_def RC.asmrec_expr_stable'_def RC.asmrec_var_stable'_def)

lemma prog_OverlayDriver_compile_succeeded:
  "let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_OverlayDriver
   in \<not> fail"
  by eval

method unstable_exprs_debug_tac =
  rule conjI |
  solves \<open>clarsimp simp:RC.asmrec_var_stable'_def\<close>

lemma prog_HidInputDriver_input_reqs:
  "RC.compile_cmd_input_reqs RC.C\<^sub>0 None 0 prog_HidInputDriver"
  unfolding RC.compile_cmd_input_reqs_def RC.C\<^sub>0_def prog_HidInputDriver_def RC.no_unstable_exprs_def
  by (force simp:RC.regrec_stable_def RC.asmrec_expr_stable'_def RC.asmrec_var_stable'_def)

lemma prog_HidInputDriver_compile_succeeded:
  "let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_HidInputDriver
   in \<not> fail"
  by eval

lemma prog_Switch_input_reqs:
  "RC.compile_cmd_input_reqs RC.C\<^sub>0 None 0 prog_Switch"
  unfolding RC.compile_cmd_input_reqs_def
  apply clarsimp
  apply(rule conjI)
   apply(force simp:prog_Switch_def)
  apply(rule conjI)
   unfolding prog_Switch_def RC.C\<^sub>0_def RC.no_unstable_exprs_def
   (* Long-running *)
   apply clarsimp
   apply(clarsimp simp:RC.lock_governed_def RC.asmrec_var_stable'_def split:lock_addr.splits)
   apply(clarsimp simp:RC.asmrec_expr_stable'_def)
   apply unstable_exprs_debug_tac+
  by (force simp:RC.regrec_stable_def)

lemma prog_Switch_compile_succeeded:
  "let (PCs, l', nl', C', fail) = RC_compile_cmd RC_C\<^sub>0 None 0 prog_Switch
   in \<not> fail"
  by eval


(* Option 1: Whole-system secure compilation theorem for this CDDC model, using Rob's
  theorems proved in RISCRefinementSafe *)

(* robs: Using newest version of theorem for jfp - one that I proved as a consequence of
  Toby's csf16 whole-system refinement theorem (modified by me to support INIT conds).
  This should not rely on any sound-mode-use facts proved directly at RISC level. *)

definition init_cs_CDDC_RISC
where
  "init_cs_CDDC_RISC \<equiv> [((0, prog_OverlayDriver_RISC), regs\<^sub>0),
                          ((0, prog_HidInputDriver_RISC), regs\<^sub>0),
                          ((0, prog_Switch_RISC), regs\<^sub>0)]"

theorem option_1a_using_theorem_from_sys_refinement:
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
  apply(rule_tac cs="map fst init_cms_CDDC"
        in RC.whole_system_compilation_sys_refine)
   apply(unfold init_cms_CDDC_RISC_def init_cms_CDDC_def)
   apply force
  apply clarsimp
  apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
   apply(erule disjE)
   apply clarsimp
    using prog_OverlayDriver_RISC_def
          prog_OverlayDriver_com_sifum_secure
          prog_OverlayDriver_locally_sound_mode_use
          prog_OverlayDriver_input_reqs
          prog_OverlayDriver_compile_succeeded
    apply force
   apply(erule disjE)
    using prog_HidInputDriver_RISC_def
          prog_HidInputDriver_com_sifum_secure
          prog_HidInputDriver_locally_sound_mode_use
          prog_HidInputDriver_input_reqs
          prog_HidInputDriver_compile_succeeded
    apply force
   using prog_Switch_RISC_def
         prog_Switch_com_sifum_secure
         prog_Switch_locally_sound_mode_use
         prog_Switch_input_reqs
         prog_Switch_compile_succeeded
   apply force
  by force

(* Using the 'whole_system_compilation_secure_thesis' theorem that was proved
  by using the compositionality theorem at RISC level. *)
(* robs: This would get broken from disabling GloballySoundRISCLockUse theory that it depends on. *)
thm RC.whole_system_compilation_secure_thesis

theorem option_1b_using_theorem_from_risc_composition:
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
  apply(rule RC.whole_system_compilation_secure_thesis)
  unfolding init_cms_CDDC_RISC_def
  apply clarsimp
  apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
   apply(erule disjE)
    using prog_OverlayDriver_RISC_def
          prog_OverlayDriver_com_sifum_secure
          prog_OverlayDriver_locally_sound_mode_use
          prog_OverlayDriver_input_reqs
          prog_OverlayDriver_compile_succeeded
    apply force
   apply(erule disjE)
    using prog_HidInputDriver_RISC_def
          prog_HidInputDriver_com_sifum_secure
          prog_HidInputDriver_locally_sound_mode_use
          prog_HidInputDriver_input_reqs
          prog_HidInputDriver_compile_succeeded
    apply force
   using prog_Switch_RISC_def
         prog_Switch_com_sifum_secure
         prog_Switch_locally_sound_mode_use
         prog_Switch_input_reqs
         prog_Switch_compile_succeeded
   apply force
  by force

(* Using the old 'whole_system_compilation_secure' theorem - consider cleaning this up. *)
(* robs: This would get broken from disabling GloballySoundRISCLockUse theory that it depends on. *)
thm RC.whole_system_compilation_secure
  RC.whole_system_compilation_secure[simplified]
  RC.compile_cmd_simple_ev\<^sub>B_eq_secure

abbreviation comp_inputs
where
   "comp_inputs \<equiv> [(RC_C\<^sub>0, None, 0), (RC_C\<^sub>0, None, 0), (RC_C\<^sub>0, None, 0)]"

theorem
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
  unfolding init_cms_CDDC_RISC_def
  apply(rule_tac comp_inputs=comp_inputs and comp_outputs=comp_outputs and cms=init_cms_CDDC
        in RC.whole_system_compilation_secure[simplified, rotated -1])
     unfolding prog_OverlayDriver_RISC_def prog_HidInputDriver_RISC_def prog_Switch_RISC_def
               init_cms_CDDC_def
     apply(force simp:Let_def split:prod.splits)
    apply force
   apply force
  apply(clarsimp simp:Let_def split:prod.splits)
  apply(rule conjI)
   using CDDC_sound_mode_use unfolding RC.WL.sound_mode_use_def init_cms_CDDC_def
   apply force
  apply(rule conjI)
   using prog_OverlayDriver_com_sifum_secure apply force
  apply(rule conjI)
   using prog_OverlayDriver_input_reqs apply blast
  apply(rule conjI)
   using prog_OverlayDriver_compile_succeeded apply force
  apply(rule conjI)
   apply clarsimp
   using CDDC_sound_mode_use unfolding RC.WL.sound_mode_use_def init_cms_CDDC_def
   apply force
  apply(rule conjI)
   using prog_HidInputDriver_com_sifum_secure apply force
  apply(rule conjI)
   using prog_HidInputDriver_input_reqs apply blast
  apply(rule conjI)
   using prog_HidInputDriver_compile_succeeded apply force
  apply(rule conjI)
   apply clarsimp
   using CDDC_sound_mode_use unfolding RC.WL.sound_mode_use_def init_cms_CDDC_def
   apply force
  apply(rule conjI)
   using prog_Switch_com_sifum_secure apply force
  apply(rule conjI)
   using prog_Switch_input_reqs apply blast
  using prog_Switch_compile_succeeded by force

(* Per-thread secure compilation theorems for this CDDC model's components *)

(* Using the old 'per_thread_compilation_secure' theorem *)
thm RC.per_thread_compilation_secure[simplified]
    RC.per_thread_compilation_secure_thesis

(* Note that the com_sifum_secure property for RISC output programs
  includes pc-security as an extra bisim requirement *)
thm RC.RL.com_sifum_secure_def
    RC.RL.low_indistinguishable_def
    RC.RL.mm_equiv.simps
    RL.bisim_pc_security_def

theorem
  "RC.RL.com_sifum_secure (((0, prog_OverlayDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  apply(rule_tac l=None and nl=0 and comp_output=comp_outputs_OverlayDriver
        in RC.per_thread_compilation_secure[simplified])
    using prog_OverlayDriver_com_sifum_secure apply blast
   unfolding prog_OverlayDriver_RISC_def
   using prog_OverlayDriver_compile_succeeded apply force
  using prog_OverlayDriver_input_reqs by blast

theorem
  "RC.RL.com_sifum_secure (((0, prog_HidInputDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  apply(rule_tac l=None and nl=0 and comp_output=comp_outputs_HidInputDriver
        in RC.per_thread_compilation_secure[simplified])
    using prog_HidInputDriver_com_sifum_secure apply blast
   unfolding prog_HidInputDriver_RISC_def
   using prog_HidInputDriver_compile_succeeded apply force
  using prog_HidInputDriver_input_reqs by blast

theorem
  "RC.RL.com_sifum_secure (((0, prog_Switch_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  apply(rule_tac l=None and nl=0 and comp_output=comp_outputs_Switch
        in RC.per_thread_compilation_secure[simplified])
    using prog_Switch_com_sifum_secure apply blast
   unfolding prog_Switch_RISC_def
   using prog_Switch_compile_succeeded apply force
  using prog_Switch_input_reqs by blast

(* Using the new 'per_thread_compilation_secure_thesis' theorem *)
thm RC.per_thread_compilation_secure_thesis'

theorem per_thread_compilation_secure_thesis'':
  "RC.WL.com_sifum_secure (c, mds\<^sub>0) \<Longrightarrow>
   \<exists> l' nl' C'. (PCs, l', nl', C', False) = (RC.compile_cmd RC_C\<^sub>0 l nl c) \<Longrightarrow>
   P = map fst PCs \<Longrightarrow>
   pc = 0 \<Longrightarrow>
   mds = mds\<^sub>0 \<Longrightarrow>
   RC.compile_cmd_input_reqs RC_C\<^sub>0 l nl c \<Longrightarrow>
   RC.RL.com_sifum_secure (((pc, P), regs), mds)"
  using RC.per_thread_compilation_secure_thesis'
  by blast

theorem prog_OverlayDriver_RISC_com_sifum_secure:
  "RC.RL.com_sifum_secure (((0, prog_OverlayDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  unfolding prog_OverlayDriver_RISC_def
  using prog_OverlayDriver_com_sifum_secure
        prog_OverlayDriver_input_reqs
        prog_OverlayDriver_compile_succeeded
  by (force intro:per_thread_compilation_secure_thesis'')

theorem prog_HidInputDriver_RISC_com_sifum_secure:
  "RC.RL.com_sifum_secure (((0, prog_HidInputDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  unfolding prog_HidInputDriver_RISC_def
  using prog_HidInputDriver_com_sifum_secure
        prog_HidInputDriver_input_reqs
        prog_HidInputDriver_compile_succeeded
  by (force intro:per_thread_compilation_secure_thesis'')

theorem prog_Switch_RISC_com_sifum_secure:
  "RC.RL.com_sifum_secure (((0, prog_Switch_RISC), regs\<^sub>0), RC.mds\<^sub>0)"
  unfolding prog_Switch_RISC_def
  using prog_Switch_com_sifum_secure
        prog_Switch_input_reqs
        prog_Switch_compile_succeeded
  by (force intro:per_thread_compilation_secure_thesis'')

thm RC.per_thread_compilation_locally_sound

theorem per_thread_compilation_locally_sound'':
  "\<exists> l' nl' C'. (PCs, l', nl', C', False) = (RC.compile_cmd RC_C\<^sub>0 l nl c) \<Longrightarrow>
   P = map fst PCs \<Longrightarrow>
   pc = 0 \<Longrightarrow>
   mds = mds\<^sub>0 \<Longrightarrow>
   RC.compile_cmd_input_reqs RC_C\<^sub>0 l nl c \<Longrightarrow>
   RL.no_locks_acquired mem \<Longrightarrow>
   RC.RL.locally_sound_mode_use \<langle>((0, P), regs), mds, mem\<rangle>\<^sub>C"
  using RC.per_thread_compilation_locally_sound'
  by blast

(* Per-thread local soundness results of compiler outputs *)

theorem prog_OverlayDriver_RISC_locally_sound:
  "RL.no_locks_acquired mem \<Longrightarrow>
   RC.RL.locally_sound_mode_use \<langle>((0, prog_OverlayDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0, mem\<rangle>\<^sub>C"
  unfolding prog_OverlayDriver_RISC_def
  using prog_OverlayDriver_com_sifum_secure
        prog_OverlayDriver_input_reqs
        prog_OverlayDriver_compile_succeeded
  by (force intro:per_thread_compilation_locally_sound'')

theorem prog_HidInputDriver_RISC_locally_sound:
  "RL.no_locks_acquired mem \<Longrightarrow>
   RC.RL.locally_sound_mode_use \<langle>((0, prog_HidInputDriver_RISC), regs\<^sub>0), RC.mds\<^sub>0, mem\<rangle>\<^sub>C"
  unfolding prog_HidInputDriver_RISC_def
  using prog_HidInputDriver_com_sifum_secure
        prog_HidInputDriver_input_reqs
        prog_HidInputDriver_compile_succeeded
  by (force intro:per_thread_compilation_locally_sound'')

theorem prog_Switch_RISC_locally_sound:
  "RL.no_locks_acquired mem \<Longrightarrow>
   RC.RL.locally_sound_mode_use \<langle>((0, prog_Switch_RISC), regs\<^sub>0), RC.mds\<^sub>0, mem\<rangle>\<^sub>C"
  unfolding prog_Switch_RISC_def
  using prog_Switch_com_sifum_secure
        prog_Switch_input_reqs
        prog_Switch_compile_succeeded
  by (force intro:per_thread_compilation_locally_sound'')


(* Option 2: Whole-system result obtained by applying compositionality theorem directly at RISC level
  to per-thread security and local soundness results, along with RISC-level global soundness. *)

theorem option_2_using_risc_composition:
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
  unfolding init_cms_CDDC_RISC_def
  apply(rule RC.RL.sifum_compositionality_cont)
   using prog_OverlayDriver_RISC_com_sifum_secure
         prog_HidInputDriver_RISC_com_sifum_secure
         prog_Switch_RISC_com_sifum_secure
   apply force
  apply clarsimp
  (* robs: This would get broken from disabling GloballySoundRISCLockUse theory that it depends on *)
  using prog_OverlayDriver_RISC_locally_sound
        prog_HidInputDriver_RISC_locally_sound
        prog_Switch_RISC_locally_sound
        RC.RL.no_locks_empty_mds_global_soundness
  by force


(* Compiled program texts obtained using Isabelle codegen execution of the compiler *)

(* For pretty-printing
declare One_nat_def [simp del]
declare One_nat_def [symmetric, simp add]
*)

value [code] "prog_OverlayDriver_RISC"

lemma prog_OverlayDriver_RISC_text:
  "prog_OverlayDriver_RISC =
   [(Some 0, MoveK r0 1),
    (None, Jz 1 r0),
    (None, InstB.LockAcq rpc_overlay_mouse_click_lock),
    (None, Load r1 rpc_overlay_mouse_click_call),
    (None, MoveK r2 0),
    (None, Op BNeq r1 r2),
    (None, Jz 2 r1),
    (None, Load r3 rpc_overlay_mouse_click_arg),
    (None, MoveK r4 212),
    (None, Op BEq r3 r4),
    (None, Jz 4 r3),
    (None, MoveK r5 0),
    (None, Store rpc_overlay_mouse_click_ret r5),
    (None, Jmp 5),
    (Some 4, Load r5 rpc_overlay_mouse_click_arg),
    (None, MoveK r6 121),
    (None, Op BEq r5 r6),
    (None, Jz 6 r5),
    (None, MoveK r7 1),
    (None, Store rpc_overlay_mouse_click_ret r7),
    (None, Jmp 7),
    (Some 6, Load r7 rpc_overlay_mouse_click_arg),
    (None, MoveK r8 50),
    (None, Op BEq r7 r8),
    (None, Jz 8 r7),
    (None, MoveK r9 2),
    (None, Store rpc_overlay_mouse_click_ret r9),
    (None, Jmp 9),
    (Some 8, MoveK r9 255),
    (None, Store rpc_overlay_mouse_click_ret r9),
    (None, Tau),
    (Some 9, Tau),
    (Some 7, Tau),
    (Some 5, MoveK r5 0),
    (None, Store rpc_overlay_mouse_click_call r5),
    (None, Jmp 3),
    (Some 2, Nop),
    (None, Tau),
    (Some 3, InstB.LockRel rpc_overlay_mouse_click_lock),
    (None, Jmp 0)]"
  by eval

value [code] "prog_HidInputDriver_RISC"

lemma prog_HidInputDriver_RISC_text:
  "prog_HidInputDriver_RISC =
   [(None, InstB.LockAcq input_driver_private_lock),
    (Some 0, MoveK r0 1),
    (None, Jz 1 r0),
    (None, InstB.LockAcq hid_read_atomicity_lock),
    (None, Load r1 hid_mouse_available),
    (None, Store temp r1),
    (None, InstB.LockRel hid_read_atomicity_lock),
    (None, MoveK r2 0),
    (None, Op BNeq r1 r2),
    (None, Jz 2 r1),
    (None, InstB.LockAcq input_event_lock),
    (None, MoveK r3 0),
    (None, Store input_event_data r3),
    (None, MoveK r4 1),
    (None, Store input_event_type r4),
    (None, InstB.LockAcq hid_read_atomicity_lock),
    (None, Load r5 hid_mouse_source),
    (None, Store input_event_data r5),
    (None, InstB.LockRel hid_read_atomicity_lock),
    (None, InstB.LockRel input_event_lock),
    (None, Jmp 3),
    (Some 2, Nop),
    (None, Tau),
    (Some 3, InstB.LockAcq hid_read_atomicity_lock),
    (None, Load r3 hid_keyboard_available),
    (None, Store temp r3),
    (None, InstB.LockRel hid_read_atomicity_lock),
    (None, MoveK r1 0),
    (None, Op BNeq r3 r1),
    (None, Jz 4 r3),
    (None, InstB.LockAcq input_event_lock),
    (None, MoveK r4 0),
    (None, Store input_event_data r4),
    (None, MoveK r5 2),
    (None, Store input_event_type r5),
    (None, Load r6 hid_keyboard_source),
    (None, Store input_event_data r6),
    (None, InstB.LockRel input_event_lock),
    (None, Jmp 5),
    (Some 4, Nop),
    (None, Tau),
    (Some 5, Jmp 0)]"
  by eval

value [code] "prog_Switch_RISC"

lemma prog_Switch_RISC_text:
  "prog_Switch_RISC =
   [(None, InstB.LockAcq switch_private_lock),
    (None, MoveK r0 0),
    (None, Store current_event_data r0),
    (None, MoveK r1 0),
    (None, Store current_event_type r1),
    (None, InstB.LockAcq input_event_lock),
    (None, MoveK r2 0),
    (None, Store input_event_data r2),
    (None, MoveK r3 0),
    (None, Store hid_keyboard_source r3),
    (None, Load r4 active_domain),
    (None, Store indicated_domain r4),
    (None, InstB.LockRel input_event_lock),
    (Some 0, MoveK r0 1),
    (None, Jz 1 r0),
    (None, InstB.LockAcq input_event_lock),
    (None, Load r1 indicated_domain),
    (None, Load r2 active_domain),
    (None, Op BEq r1 r2),
    (None, Jz 2 r1),
    (None, Load r3 input_event_type),
    (None, Store current_event_type r3),
    (None, Load r4 input_event_data),
    (None, Store current_event_data r4),
    (None, Jmp 3),
    (Some 2, Nop),
    (None, Tau),
    (Some 3, InstB.LockRel input_event_lock),
    (None, Load r1 current_event_type),
    (None, MoveK r3 1),
    (None, Op BEq r1 r3),
    (None, Jz 4 r1),
    (None, InstB.LockAcq rpc_overlay_mouse_click_lock),
    (None, Load r4 current_event_data),
    (None, Store rpc_overlay_mouse_click_arg r4),
    (None, MoveK r5 1),
    (None, Store rpc_overlay_mouse_click_call r5),
    (None, InstB.LockRel rpc_overlay_mouse_click_lock),
    (None, MoveK r4 0),
    (None, Store done_rpc r4),
    (Some 6, Load r0 done_rpc),
    (None, MoveK r1 0),
    (None, Op BNeq r0 r1),
    (None, Jz 7 r0),
    (None, InstB.LockAcq rpc_overlay_mouse_click_lock),
    (None, Load r2 rpc_overlay_mouse_click_call),
    (None, MoveK r3 0),
    (None, Op BNeq r2 r3),
    (None, Jz 8 r2),
    (None, Load r4 rpc_overlay_mouse_click_ret),
    (None, Store overlay_result r4),
    (None, MoveK r5 1),
    (None, Store done_rpc r5),
    (None, Jmp 9),
    (Some 8, Nop),
    (None, Tau),
    (Some 9, InstB.LockRel rpc_overlay_mouse_click_lock),
    (None, Jmp 6),
    (Some 7, Load r0 overlay_result),
    (None, MoveK r1 255),
    (None, Op BNeq r0 r1),
    (None, Jz 10 r0),
    (None, MoveK r2 2),
    (None, Store cursor_domain r2),
    (None, Jmp 11),
    (Some 10, Load r2 current_event_data),
    (None, Store compositor_cursor_position r2),
    (None, InstB.LockAcq compositor_read_atomicity_lock),
    (None, Load r3 compositor_domain_under_cursor),
    (None, Store cursor_domain r3),
    (None, InstB.LockRel compositor_read_atomicity_lock),
    (None, MoveK r4 255),
    (None, Op BEq r3 r4),
    (None, Jz 12 r3),
    (None, Load r5 active_domain),
    (None, Store cursor_domain r5),
    (None, Jmp 13),
    (Some 12, Nop),
    (None, Tau),
    (Some 13, Tau),
    (Some 11, Load r2 cursor_domain),
    (None, MoveK r3 2),
    (None, Op BEq r2 r3),
    (None, Jz 14 r2),
    (None, Load r4 overlay_result),
    (None, MoveK r5 2),
    (None, Op BNeq r4 r5),
    (None, Load r6 switch_state_mouse_down),
    (None, MoveK r7 0),
    (None, Op BEq r6 r7),
    (None, Load r8 current_event_data),
    (None, MoveK r9 42),
    (None, Op BEq r8 r9),
    (None, Load r10 overlay_result),
    (None, MoveK r11 255),
    (None, Op BNeq r10 r11),
    (None, Load r12 overlay_result),
    (None, Load r13 active_domain),
    (None, Op BNeq r12 r13),
    (None, Op BAnd r10 r12),
    (None, Op BAnd r8 r10),
    (None, Op BAnd r6 r8),
    (None, Op BAnd r4 r6),
    (None, Jz 16 r4),
    (None, Load r14 overlay_result),
    (None, Store active_domain r14),
    (None, InstB.LockAcq input_event_lock),
    (None, MoveK r4 0),
    (None, Store input_event_data r4),
    (None, MoveK r6 0),
    (None, Store input_event_type r6),
    (None, MoveK r8 0),
    (None, Store hid_keyboard_source r8),
    (None, Store indicated_domain r14),
    (None, InstB.LockRel input_event_lock),
    (None, Jmp 17),
    (Some 16, Nop),
    (None, Tau),
    (Some 17, Jmp 15),
    (Some 14, Load r4 switch_state_mouse_down),
    (None, MoveK r5 0),
    (None, Op BEq r4 r5),
    (None, Load r6 current_event_data),
    (None, MoveK r7 42),
    (None, Op BEq r6 r7),
    (None, Load r8 active_domain),
    (None, Load r9 cursor_domain),
    (None, Op BNeq r8 r9),
    (None, Op BAnd r6 r8),
    (None, Op BAnd r4 r6),
    (None, Jz 18 r4),
    (None, Store active_domain r9),
    (None, InstB.LockAcq input_event_lock),
    (None, MoveK r4 0),
    (None, Store input_event_data r4),
    (None, MoveK r6 0),
    (None, Store input_event_type r6),
    (None, MoveK r8 0),
    (None, Store hid_keyboard_source r8),
    (None, Store indicated_domain r9),
    (None, InstB.LockRel input_event_lock),
    (None, Jmp 19),
    (Some 18, Nop),
    (None, Tau),
    (Some 19, Load r4 switch_state_mouse_down),
    (None, MoveK r6 0),
    (None, Op BNeq r4 r6),
    (None, Load r8 current_event_data),
    (None, MoveK r9 42),
    (None, Op BEq r8 r9),
    (None, Op BOr r4 r8),
    (None, Jz 20 r4),
    (None, Load r10 active_domain),
    (None, MoveK r11 0),
    (None, Op BEq r10 r11),
    (None, Jz 22 r10),
    (None, Load r12 current_event_data),
    (None, Store output_event_buffer0 r12),
    (None, Jmp 23),
    (Some 22, Load r12 current_event_data),
    (None, Store output_event_buffer1 r12),
    (None, Tau),
    (Some 23, Jmp 21),
    (Some 20, Load r10 cursor_domain),
    (None, MoveK r11 0),
    (None, Op BEq r10 r11),
    (None, Jz 24 r10),
    (None, Load r12 current_event_data),
    (None, Store output_event_buffer0 r12),
    (None, Jmp 25),
    (Some 24, Load r12 current_event_data),
    (None, Store output_event_buffer1 r12),
    (None, Tau),
    (Some 25, Tau),
    (Some 21, Tau),
    (Some 15, Load r4 current_event_data),
    (None, MoveK r5 42),
    (None, Op BEq r4 r5),
    (None, Jz 26 r4),
    (None, MoveK r6 1),
    (None, Store switch_state_mouse_down r6),
    (None, Jmp 27),
    (Some 26, MoveK r6 0),
    (None, Store switch_state_mouse_down r6),
    (None, Tau),
    (Some 27, Jmp 5),
    (Some 4, Nop),
    (None, Tau),
    (Some 5, Load r0 current_event_type),
    (None, MoveK r1 2),
    (None, Op BEq r0 r1),
    (None, Jz 28 r0),
    (None, Load r2 active_domain),
    (None, MoveK r3 0),
    (None, Op BEq r2 r3),
    (None, Jz 30 r2),
    (None, Load r4 current_event_data),
    (None, Store output_event_buffer0 r4),
    (None, Jmp 31),
    (Some 30, Load r4 current_event_data),
    (None, Store output_event_buffer1 r4),
    (None, Tau),
    (Some 31, Jmp 29),
    (Some 28, Nop),
    (None, Tau),
    (Some 29, MoveK r2 0),
    (None, Store current_event_data r2),
    (None, MoveK r3 0),
    (None, Store current_event_type r3),
    (None, Jmp 0)]"
  by eval


(* Option 3: Whole-system secure compilation theorem for this CDDC model,
  using Toby's whole-system locale and refined_prog_secure lemma *)

lemma [intro]:
  "\<And>mem. RC.WL.pred P\<^sub>0 mem"
  unfolding P\<^sub>0_def
  by simp

(* robs: Here I used the version I simplified for presentation in the jfp paper. *)

theorem option_3_using_sys_refinement:
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
proof -
  interpret SRSMI: sifum_refinement_same_mem_init dma \<C>_vars \<C> WL.eval\<^sub>w RL.eval\<^sub>r
                   1 WL.no_locks_acquired RL.no_locks_acquired WL.bisim_simple_ev\<^sub>B_eq
    by (unfold_locales)
  show ?thesis
    apply(rule_tac \<B>s="[TS.\<R> \<Gamma>\<^sub>0 \<S>\<^sub>0 P\<^sub>0,
                        TS.\<R> \<Gamma>'_HidInputDriver \<S>'_HidInputDriver P'_HidInputDriver,
                        TS.\<R> \<Gamma>'_Switch \<S>'_Switch P'_Switch]" and
                   \<R>s="replicate 3 RC.\<R>" and
                   \<I>s="replicate 3 RC.\<I>" and
                   cms\<^sub>A="init_cms_CDDC"
          in SRSMI.refined_prog_secure_init_simplified_for_print)
    apply(rule conjI)
     using RC.WL_RL_no_locks_acquired apply blast
    apply(rule conjI)
     (* That the original While-language program satisfies
        its compositionality requirements. We already proved this. *)
     using CDDC_sound_mode_use
     apply force
    unfolding init_cms_CDDC_def init_cms_CDDC_RISC_def
    apply clarsimp
    (* Properties for each thread's refinement *)
    apply(case_tac i)
     (* OverlayDriver *)
     apply clarsimp
     apply(rule conjI)
      apply(clarsimp simp:SRSMI.compositional_refinement_def)
      apply(rule conjI)
       (* That \<B>, \<R>, \<I> together form a secure refinement *)
       using RC.secure_refinement_\<R>_simple_ev\<^sub>B_eq_\<B>[simplified]
             TS.\<R>_bisim TS.bisim_simple_ev\<^sub>B_eq_\<R>\<^sub>u
       apply blast
      apply(rule conjI)
      (* That \<B> is a strong low-bisimulation modulo modes *)
       using TS.\<R>_bisim
       apply blast
      apply(rule conjI)
       (* That \<R> preserves local guarantee compliance *)
       apply(rule RC.establishes_local_guarantee_compliance_\<R>[simplified])
      apply(rule conjI)
       (* That \<I> is symmetric *)
       apply(force simp:sym_def RC.\<I>.simps)
      (* That all low-equal memories that satisfy INIT\<^sub>r are paired by \<I> *)
      apply(force simp:RC.\<I>.simps)
     (* That we're initially in \<R>. The compiler gives us that. *)
     apply(rule conjI)
      apply(clarsimp simp:prog_OverlayDriver_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
      apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_OverlayDriver_Cs])
         apply clarsimp
         apply(rule conjI)
          apply(force simp:prog_OverlayDriver_Cs_def zip_map_fst_snd)
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         using prog_OverlayDriver_compile_succeeded
         apply force
        apply(force simp:prog_OverlayDriver_input_reqs)
       apply(force simp:prog_OverlayDriver_Cs_def)
      apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
     (* That we're initially in \<B>. The security type system gives us that. *)
     apply clarsimp
     using prog_OverlayDriver_typed
     apply(rule TS.Typed_in_\<R>)
         apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def P\<^sub>0_def)
        apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        using TS.type_max_dma_type apply force
       apply blast
      apply blast
     apply(force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
    (* Rinse and repeat for the other two components.
      Proofs are much the same for this system because they are all obtained
      using the same security type system and compiler. *)
    apply(case_tac "i = 1")
     (* HidInputDriver *)
     apply clarsimp
     apply(rule conjI)
      apply(clarsimp simp:SRSMI.compositional_refinement_def)
      apply(rule conjI)
       (* That \<B>, \<R>, \<I> together form a secure refinement *)
       using RC.secure_refinement_\<R>_simple_ev\<^sub>B_eq_\<B>[simplified]
             TS.\<R>_bisim TS.bisim_simple_ev\<^sub>B_eq_\<R>\<^sub>u
       apply blast
      apply(rule conjI)
      (* That \<B> is a strong low-bisimulation modulo modes *)
       using TS.\<R>_bisim
       apply blast
      apply(rule conjI)
       (* That \<R> preserves local guarantee compliance *)
       apply(rule RC.establishes_local_guarantee_compliance_\<R>[simplified])
      apply(rule conjI)
       (* That \<I> is symmetric *)
       apply(force simp:sym_def RC.\<I>.simps)
      (* That all low-equal memories that satisfy INIT\<^sub>r are paired by \<I> *)
      apply(force simp:RC.\<I>.simps)
     (* That we're initially in \<R>. The compiler gives us that. *)
     apply(rule conjI)
      apply(clarsimp simp:prog_HidInputDriver_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
      apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_HidInputDriver_Cs])
         apply clarsimp
         apply(rule conjI)
          apply(force simp:prog_HidInputDriver_Cs_def zip_map_fst_snd)
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         using prog_HidInputDriver_compile_succeeded
         apply force
        apply(force simp:prog_HidInputDriver_input_reqs)
       apply(force simp:prog_HidInputDriver_Cs_def)
      apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
     (* That we're initially in \<B>. The security type system gives us that. *)
     apply clarsimp
     using prog_HidInputDriver_typed
     apply clarsimp
     apply(rule TS.Typed_in_\<R>)
          apply force
         apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def P\<^sub>0_def)
        apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        using TS.type_max_dma_type apply force
       apply blast
      apply blast
     apply(force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
    (* Switch *)
    apply clarsimp
    apply(rule conjI)
     apply(clarsimp simp:SRSMI.compositional_refinement_def)
     apply(rule conjI)
      (* That \<B>, \<R>, \<I> together form a secure refinement *)
      using RC.secure_refinement_\<R>_simple_ev\<^sub>B_eq_\<B>[simplified]
            TS.\<R>_bisim TS.bisim_simple_ev\<^sub>B_eq_\<R>\<^sub>u
      apply blast
     apply(rule conjI)
     (* That \<B> is a strong low-bisimulation modulo modes *)
      using TS.\<R>_bisim
      apply blast
     apply(rule conjI)
      (* That \<R> preserves local guarantee compliance *)
      apply(rule RC.establishes_local_guarantee_compliance_\<R>[simplified])
     apply(rule conjI)
      (* That \<I> is symmetric *)
      apply(force simp:sym_def RC.\<I>.simps)
     (* That all low-equal memories that satisfy INIT\<^sub>r are paired by \<I> *)
     apply(force simp:RC.\<I>.simps)
    (* That we're initially in \<R>. The compiler gives us that. *)
    apply(rule conjI)
     apply(clarsimp simp:prog_Switch_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
     apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_Switch_Cs])
        apply clarsimp
        apply(rule conjI)
         apply(force simp:prog_Switch_Cs_def zip_map_fst_snd)
        apply(rule conjI)
         apply force
        apply(rule conjI)
         apply force
        apply(rule conjI)
         apply force
        using prog_Switch_compile_succeeded
        apply force
       apply(force simp:prog_Switch_input_reqs)
      apply(force simp:prog_Switch_Cs_def)
     apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
    (* That we're initially in \<B>. The security type system gives us that. *)
    apply clarsimp
    using prog_Switch_typed
    apply clarsimp
    apply(rule TS.Typed_in_\<R>)
         apply force
        apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def P\<^sub>0_def)
       apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
       using TS.type_max_dma_type apply force
      apply blast
     apply blast
    by (force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
qed

(* robs: Here I used Toby's original locale and lemma directly,
         not the ones I simplified for print. *)

(* A "cms" according to sifum_refinement_sys is a list of these:
  record ('a, 'Val, 'Var\<^sub>C, 'Com\<^sub>C, 'Var\<^sub>A, 'Com\<^sub>A) componentwise_refinement =
    priv_mem :: "'Var\<^sub>C set" (* private variables *)
    \<R>\<^sub>A_rel :: "('Com\<^sub>A, 'Var\<^sub>A, 'Val) LocalConf rel" (* abstract bisimulation *)
    \<R>_rel :: "('Com\<^sub>A, 'Var\<^sub>A, 'Val, 'Com\<^sub>C, 'Var\<^sub>C) state_relation" (* refinement relation *)
    P_rel :: "('Com\<^sub>C, 'Var\<^sub>C, 'Val) LocalConf rel"
  To disambiguate from the "cms" I refer to in RISCRefinementSafe, I will use "crs" here.
*)
definition my_crs
where
  "my_crs \<equiv> [
     \<lparr> priv_mem = {},
       \<R>\<^sub>A_rel = TS.\<R> \<Gamma>\<^sub>0 \<S>\<^sub>0 P\<^sub>0,
       \<R>_rel = RC.\<R>,
       P_rel = RC.\<I> \<rparr>,
     \<lparr> priv_mem = {},
       \<R>\<^sub>A_rel = TS.\<R> \<Gamma>'_HidInputDriver \<S>'_HidInputDriver P'_HidInputDriver,
       \<R>_rel = RC.\<R>,
       P_rel = RC.\<I> \<rparr>,
     \<lparr> priv_mem = {},
       \<R>\<^sub>A_rel = TS.\<R> \<Gamma>'_Switch \<S>'_Switch P'_Switch,
       \<R>_rel = RC.\<R>,
       P_rel = RC.\<I> \<rparr>
     ]"

(* Wishlist: a function that returns the \<B> of a typecheck, and these facts about that \<B>.
    TM.strong_low_bisim_mm ?\<B> \<Longrightarrow>
    RC.WL.bisim_simple_ev\<^sub>B_eq ?\<B>
  The typecheck currently requires user intervention, so I'm not sure we can just write
  a function that does this.
  But that might be nice to think about for future work. *)

(* TODO: push these up? *)
lemma \<I>_sym':
  "(\<langle>(c, regs), mds, mem\<rangle>\<^sub>C, \<langle>(c', regs'), mds', mem'\<rangle>\<^sub>C) \<in> RC.\<I> \<Longrightarrow>
   (\<langle>(c', regs'), mds', mem'\<rangle>\<^sub>C, \<langle>(c, regs), mds, mem\<rangle>\<^sub>C) \<in> RC.\<I>"
  apply(induct rule:RC.\<I>.induct)
  using RC.\<I>.\<I>_default by blast

lemma \<I>_sym: "sym RC.\<I>"
  unfolding sym_def
  using \<I>_sym' by force

interpretation
  SRSI: sifum_refinement_sys_init dma dma \<C>_vars \<C>_vars \<C> \<C> WL.eval\<^sub>w RL.eval\<^sub>r undefined WL.no_locks_acquired RL.no_locks_acquired "(\<lambda>_. True)" id my_crs "[{}, {}, {}]"
  apply(unfold_locales)
          apply(force simp:my_crs_def nth_Cons')
         apply(clarsimp simp:my_crs_def)
         apply(rule set_eqI)
         apply(solves \<open>(rename_tac x, case_tac x, clarsimp, rename_tac y, case_tac y, force+)+\<close>)
        apply fastforce
       apply(clarsimp simp:my_crs_def)
       apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
        apply(force simp:RC.secure_refinement_\<R>_simple_ev\<^sub>B_eq_\<B> TS.\<R>_bisim TS.bisim_simple_ev\<^sub>B_eq_\<R>\<^sub>u)
       apply linarith
      apply(clarsimp simp:my_crs_def)
      apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
       apply(force simp:RC.establishes_local_guarantee_compliance_\<R>)
       apply linarith
     apply(clarsimp simp:my_crs_def)
     apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
      apply(force simp:TS.\<R>_bisim)
     apply linarith
    apply(clarsimp simp:my_crs_def)
    apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
     apply(force simp:\<I>_sym)
    apply linarith
   apply(clarsimp simp:my_crs_def)
   apply(case_tac "i = 0 \<or> i = 1 \<or> i = 2")
    apply(force simp:RC.\<I>.\<I>_default)
   apply linarith
  by (force simp:my_crs_def)

declare P\<^sub>0_def [simp]

thm SRSI.refined_prog_secure_init[where cms\<^sub>A=init_cms_CDDC]

theorem
  "RC.RL.prog_sifum_secure_cont init_cms_CDDC_RISC"
  apply(rule SRSI.refined_prog_secure_init[where cms\<^sub>A=init_cms_CDDC])
        apply(force simp:RL.no_locks_acquired_def RC.WL.no_locks_acquired_def RC.mem\<^sub>A_of_def)
       apply(force simp:init_cms_CDDC_RISC_def my_crs_def)
      apply(force simp:init_cms_CDDC_def my_crs_def)
     apply(clarsimp simp:init_cms_CDDC_def init_cms_CDDC_RISC_def my_crs_def)
     apply(case_tac i)
      (* OverlayDriver *)
      apply(clarsimp simp:prog_OverlayDriver_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
      apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_OverlayDriver_Cs])
         apply clarsimp
         apply(rule conjI)
          apply(force simp:prog_OverlayDriver_Cs_def zip_map_fst_snd)
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         using prog_OverlayDriver_compile_succeeded
         apply force
        apply(force simp:prog_OverlayDriver_input_reqs)
       apply(force simp:prog_OverlayDriver_Cs_def)
      apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
     (* HidInputDriver *)
     apply(case_tac "i = 1")
      apply(clarsimp simp:prog_HidInputDriver_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
      apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_HidInputDriver_Cs])
         apply clarsimp
         apply(rule conjI)
          apply(force simp:prog_HidInputDriver_Cs_def zip_map_fst_snd)
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         apply(rule conjI)
          apply force
         using prog_HidInputDriver_compile_succeeded
         apply force
        apply(force simp:prog_HidInputDriver_input_reqs)
       apply(force simp:prog_HidInputDriver_Cs_def)
      apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
     (* Switch *)
     apply(clarsimp simp:prog_Switch_RISC_def RC.mem\<^sub>A_of_def split:prod.splits)
     apply(rule RC.compile_cmd_correctness_\<R>'[where C=RC_C\<^sub>0 and l=None and nl=0 and Cs=prog_Switch_Cs])
        apply clarsimp
        apply(rule conjI)
         apply(force simp:prog_Switch_Cs_def zip_map_fst_snd)
        apply(rule conjI)
         apply force
        apply(rule conjI)
         apply force
        apply(rule conjI)
         apply force
        using prog_Switch_compile_succeeded
        apply force
       apply(force simp:prog_Switch_input_reqs)
      apply(force simp:prog_Switch_Cs_def)
     apply(force intro:RC.no_locks_acquired_C\<^sub>0_mds\<^sub>0_init_reqs)
    apply(clarsimp simp:init_cms_CDDC_def init_cms_CDDC_RISC_def my_crs_def)
    apply(case_tac i)
     (* OverlayDriver *)
     apply clarsimp
     using prog_OverlayDriver_typed
     apply(clarsimp simp:RC.mem\<^sub>A_of_def)
     apply(rule TS.Typed_in_\<R>)
          apply force
         apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        using TS.type_max_dma_type apply force
       apply force
      apply force
     apply(force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
    apply(case_tac "i = 1")
     (* HidInputDriver *)
     apply clarsimp
     using prog_HidInputDriver_typed
     apply(clarsimp simp:RC.mem\<^sub>A_of_def)
     apply(rule TS.Typed_in_\<R>)
          apply force
         apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
        using TS.type_max_dma_type apply force
       apply force
      apply force
     apply(force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
    (* Switch *)
    apply clarsimp
    using prog_Switch_typed
    apply(clarsimp simp:RC.mem\<^sub>A_of_def)
    apply(rule TS.Typed_in_\<R>)
         apply force
        apply(force simp:TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
       apply(clarsimp simp:TM.low_mds_eq_def TS.\<Gamma>_of_mds_def TS.\<S>_of_mds_def empty_mds_def)
       using TS.type_max_dma_type apply force
      apply force
     apply force
    apply(force simp:TS.\<Gamma>_of_mds_def empty_mds_def)
   apply(rule CDDC_sound_mode_use)
   apply force
  apply(clarsimp simp:SRSI.modes_respect_priv_def init_cms_CDDC_RISC_def
          SRSI.priv_is_asm_priv_def SRSI.priv_is_guar_priv_def
          SRSI.new_asms_only_for_priv_def SRSI.new_asms_NoReadOrWrite_only_def)
  apply(rule conjI)
   apply(clarsimp simp:my_crs_def)
   apply(metis empty_iff less_Suc0 less_antisym nth_Cons_0 nth_Cons_Suc)
  apply(rule conjI)
   apply(clarsimp simp:my_crs_def)
   apply(metis empty_iff less_Suc0 less_antisym nth_Cons_0 nth_Cons_Suc)
  apply(rule conjI)
   apply(clarsimp simp:my_crs_def)
   apply(metis LockVars.empty_mds_no_asms less_Suc0 less_antisym nth_Cons_0 nth_Cons_Suc)
  apply(clarsimp simp:my_crs_def)
  by (simp add:LockVars.empty_mds_no_asms nth_Cons')

end
